require("lib.tiles")
vector = require("hump.vector")
-- Function found on the internet for drawing an arc based on a start angle
-- and a sweep.  Avoids problems with start/stop sweep angles.
function arc (mode, x, y, radius, start, sweep, segments, omitlines)
   local points = { }
   
   sweep = math.max(-math.pi * 2, math.min(sweep, math.pi * 2))
   
   local o = 0
   if mode == "fill" or not omitlines then
      o = 2
      points[1] = x
      points[2] = y
   end
   
   segments = (segments or 10) - 1
   
   for i = 0, segments do
      local dir = start + (i / segments) * sweep
      
      points[o+i*2+1] = math.cos(dir) * radius + x
      points[o+i*2+2] = -math.sin(dir) * radius + y
   end

   if mode == "fill" then
      love.graphics.polygon("fill", points)
   elseif mode == "line" then
      if not omitlines then
         points[#points+1] = x
         points[#points+1] = y
      end
      love.graphics.line(points)
   end
end

-- Function that draws the tiles themselves.
-- They are basically just rectangles.
function draw_tile(tile_object)
   local tile_idx = tile_object.idx
   -- Get the coordinates of the tile based on its index.
   local pixel_coords = tile_idx_to_pixels(tile_idx)
   local idx_pos = tile_object.pos_idx

   -- Draw the box around the tile.
   love.graphics.rectangle("line",pixel_coords.x,pixel_coords.y,tile_size.x,tile_size.y)

   -- Print the tile index on the tile.
   love.graphics.print(tile_idx,((pixel_coords+tile_size/2)+vector(0,-20)):unpack())

   -- Print the "coordinate" of the tile (1-3),(1-3)
   love.graphics.print(tostring(idx_pos),(pixel_coords+tile_size/2):unpack())

   -- Print the connections on this tile.
   --love.graphics.print(to_string(tile),pixel_coords[1],pixel_coords[2])

   -- Now draw the connections present on this tile.
   for i,c in pairs(tile_object.connections) do
      draw_connection(tile_idx,{i,c})
   end
end

-- This function draws the various kinds of connections
-- by calling more specialized functions.
function draw_connection(tile_idx,connection)
   local tile_coords = tile_idx_to_pixels(tile_idx)
   if connection[1] == connection[2] then
      -- A connection from an edge to itself.
      draw_connection_same(connection,tile_coords)
   elseif table.contains({1,3},math.abs((connection[1])-(connection[2]))) then
      -- Adjacent edges, draw a large curve centered on the tile corner.
      draw_connection_adjacent(connection,tile_coords)
   else
      -- Edges on opposite sides of a tile connected
      draw_connection_across(connection,tile_coords)
   end
end

-- This function is used to draw a connection between
-- an edge and itself (e.g. "22").
function draw_connection_same(connection,tile_coords)
   assert(connection[1] == connection[2],"Trying to draw same-edge connection but the connection is different.")
   local con = connection[1] -- Equal to connection[2], just use one of them.
   local offset = -1
   local start,sweep = con*math.pi/2, math.pi
   if con % 2 == 0 then
      offset = tile_size.x/2
   else
      offset = tile_size.y/2
   end
   local center = edge_pos_to_tile_pos(con,offset)
   local pixel_pos = center+tile_coords
   
   arc("line",pixel_pos.x,pixel_pos.y,offset/2,start,sweep,20)
end

-- This function draws a connection between adjacent edges
function draw_connection_adjacent(connection,tile_coords)
   local c1,c2 = connection[1],connection[2]
   local center
   local start,sweep = 0,math.pi/2
   -- There are two radii, depending on the direction of the edge.
   local r1,r2 = tile_size.x/4,3*tile_size.y/4
   local radius = 0
   if table.contains(connection,1) and table.contains(connection,2) then
      center = vector(tile_size.x,0)
      start = math.pi
      if c1 == 1 then radius = r2 else radius = r1 end
   elseif table.contains(connection,2) and table.contains(connection,3) then
      center = vector(0,0)
      start = 3*math.pi/2
      if c1 == 2 then radius = r2 else radius = r1 end
   elseif table.contains(connection,3) and table.contains(connection,4) then
      center = vector(0,tile_size.y)
      start = 0
      if c1 == 3 then radius = r2 else radius = r1 end
   elseif table.contains(connection,4) and table.contains(connection,1) then
      center = tile_size
      start = math.pi/2
      if c1 == 4 then radius = r2 else radius = r1 end
   end
   -- arc("line",center.x+tile_coords.x,center.y+tile_coords.y,radius,
   --     start,sweep)
   local pixel_pos = center+tile_coords
   arc("line",pixel_pos.x,pixel_pos.y,radius,
       start,sweep)
end

-- This function draws a connection between edges across
-- a tile.
function draw_connection_across(connection,tile_coords)
   local quarter_h, threequarter_h = tile_size.x/4, 3*tile_size.x/4
   local quarter_v, threequarter_v = tile_size.y/4, 3*tile_size.y/4
   local start_offset, stop_offset = -1, -1
   if connection[1] == 1 then
      start_offset = threequarter_v
   elseif connection[1] == 2 then
      start_offset = threequarter_h
   elseif connection[1] == 3 then
      start_offset = quarter_v
   else
      start_offset = quarter_h
   end

   if connection[2] == 1 then
      stop_offset = quarter_v
   elseif connection[2] == 2 then
      stop_offset = quarter_h
   elseif connection[2] == 3 then
      stop_offset = threequarter_v
   else
      stop_offset = threequarter_h
   end

   local start = edge_pos_to_tile_pos(connection[1],start_offset)
   local stop = edge_pos_to_tile_pos(connection[2],stop_offset)
   local start_board = start+tile_coords
   local stop_board = stop+tile_coords
   love.graphics.line(start_board.x,start_board.y,
		      stop_board.x,stop_board.y) 
   -- local third_point
   -- if start[1] == stop[1] then
   --    -- Vertical line
   --    if start[1] < tile_size[1]/2 then
   -- 	 -- On left side of tile
   -- 	 third_point = {3*tile_size[1]/8,tile_size[2]/2}
   --    else
   -- 	 -- On right side of tile
   -- 	 third_point = {5*tile_size[1]/8,tile_size[2]/2}
   --    end
   -- else
   --    -- Horizontal line
   --    if start[2] < tile_size[2]/2 then
   -- 	 -- On upper side of tile
   -- 	 third_point = {tile_size[1]/2,3*tile_size[2]/8}
   --    else
   -- 	 -- On lower side of tile
   -- 	 third_point = {tile_size[1]/2,5*tile_size[2]/8}
   --    end
   -- end
   --draw_arcfromthree({start[1]+tile_coords[1],start[2]+tile_coords[2]},
--			{tile_size[1]/2+tile_coords[1],tile_size[2]/2+tile_coords[2]},
--			{stop[1]+tile_coords[1],stop[2]+tile_coords[2]})

end

-- Draws a circle inscribed by three points.
-- Fails silently if the points are colinear.
function draw_arcfromthree(A,B,C)
   -- A,B,C are vectors
   -- local Ax,Ay = A[1],A[2]
   -- local Bx,By = B[1],B[2]
   -- local Cx,Cy = C[1],C[2]
   local D = 2*(A.x*(B.y-C.y)+ B.x*(C.y-A.y) + C.x*(A.y-B.y))
   
   -- local Ux = ( (Ax^2+Ay^2)*(By-Cy) + (Bx^2+By^2)*(Cy-Ay) + (Cx^2+Cy^2)*(Ay-By))/D
   -- local Uy = ( (Ax^2+Ay^2)*(Cx-Bx) + (Bx^2+By^2)*(Ax-Cx) + (Cx^2+Cy^2)*(Bx-Ax))/D
   -- local R = math.sqrt((Ux-Ax)^2+(Uy-Ay)^2)
   local U = vector(0,0)
   U.x = ( A:len2()*(B.y-C.y) + B:len2()*(C.y-A.y) + C:len2()*(A.y-B.y))/D
   U.y = ( A:len2()*(C.x-B.x) + B:len2()*(A.x-C.x) + C:len2()*(B.x-A.x))/D

   local R = (U-A):len()

   local start = -math.atan2(A.y-U.y,A.x-U.x)
   --local middle = -math.atan2(By-Uy,Bx-Ux)
   local stop = -math.atan2(C.y-U.y,C.x-U.x)
   if math.abs(stop-start) > math.pi then
      stop,start = start,stop
   end
   arc("line",U.x,U.y,R,start,stop-start,20,true)
   --arc("line",Ux,Uy,R,middle,stop-middle,20,true)
end

-- This function takes a distance along an edge number (1-4)
-- and returns the pixel offset of that position relative to
-- the tile.
function edge_pos_to_tile_pos(edge_number,pos)
   local tile_pos = vector(0,0)
   if edge_number == 1 then
      tile_pos = vector(tile_size.x,pos)
   elseif edge_number == 2 then
      tile_pos.x = pos
   elseif edge_number == 3 then
      tile_pos.y = pos
   else
      tile_pos = vector(pos,tile_size.y)
   end
   return tile_pos
end


-- This function returns true/false depending on
-- whether element is contained in table.
function table.contains(table, element)
  for _, value in pairs(table) do
    if value == element then
      return true
    end
  end
  return false
end

-- This function takes a tile index (1-9) and 
-- returns a table with the coordinate on the board
-- of that tile, in pixels.
function tile_idx_to_pixels(tile_idx)
   local idx_pos = tile_idx_to_idx_pos(tile_idx)
   return (idx_pos-vector(1,1)):permul(tile_size)
end
