class = require("hump.class")
vector = require("hump.vector")
require "lib.table_print"
-- This file defines all the possible tiles, creates a random board, 
-- and makes all the connections.

-- My tiles are rectangular and have an edge convention.
-- Edge 1: right, 2: top, 3: right, 4: bottom

-- This function "rotates" the connections on a tile,
-- doing 1->2->3->4->1.
local function rotate_tile_once(tile)
   for i,connection in ipairs(tile) do
      for p = 1,2 do
	 connection[p] = (connection[p] % 4) + 1
      end
   end
end

-- This function applies the rotation function on a tile multiple times
local function rotate_tile(n,tile)
   for i=1,n,1 do
      rotate_tile_once(tile)
   end
end

-- This function copies tables recursively.
local function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

-- This table of strings contains a short definition of all the types of tiles
-- that are in my game.  Each string represents a tile, each integer represents
-- an edge on that tile.  The integers come in pairs, the first being the 
-- edge where the connection comes in, the second being the edge where the connection
-- leaves the tile.  Two tile types have two edges, 5 have three, and more have four.
local tiles_short = {"11,22","12,21",
		     "42,14,21","22,41,14","44,21,12","22,44,11","42,24,11",
		     "11,22,33,44","23,32,44,11","22,31,13,44","21,32,13,44",
		     "14,41,22,33","12,21,34,43","13,31,24,42","14,43,32,21"}
-- This table will be filled with a table-of-tables-of-tables for all the unique
-- tile types.
local tiles = {}

-- Iterate over the unique tile types.
for _,short in ipairs(tiles_short) do
   local connections = {}
   -- Iterate over integer pairs
   for connection in short:gmatch("%d%d") do
      assert(#connection == 2, "Connections only have two ends.")
      local start,stop = connection:sub(1,1), connection:sub(2,2)
      -- The subtable contains the start and stop edges for that connection
      -- for that tile.
      table.insert(connections,{tonumber(start),tonumber(stop)})
   end
   table.insert(tiles,connections)
end

-- The board size could in principle be changed, but 3,3 is the smallest possible.
local min_size = vector(3,3)
-- I haven't actually planned to change the board size, likely things will break.
board_size = vector(3,3)
n_tiles = board_size.x*board_size.y
assert(min_size <= board_size,"Board size is below the minimum required.")
-- Vector-wise <= means a.x <= b.x and a.y <= b.y

local edges,vert_edge,horiz_edge = 0,0,0
local tile_edges = -1
local r_tile = -1
local rotations = 0

-- Board tiles are filled row-by-row.
-- 1 2 3 (for 3x3) 
-- 4 5 6
-- 7 8 9
-- So tile number N is in position {N % board_size[1]+1, math.floor(N/board_size[2])}
-- The corner tiles are all defined with connections between edges 1 and 2
-- The edge tiles are all defined with connections between edges 1, 2, and 4
-- So the corner and side tiles will need to be rotated.  The center tile does not need to be rotated, but
-- can be rotated by a random amount.

-- This table will be filled with the actual tiles used for this game.
board_tiles = {}

math.randomseed(os.time())

-- Iterate over rows first, columns in the inner loop for [x][y] indexing.
-- At each step I see if the tile is at the edge of the board.
for row = 1,board_size.y do
   if row == 1 or row == board_size.y then
      horiz_edge = 1
   else
      horiz_edge = 0
   end
   for col = 1,board_size.x do
      if col == 1 or col == board_size.x then
	 vert_edge = 1
      else
	 vert_edge = 0
      end
      edges = vert_edge+horiz_edge
      -- Draw a random tile until one with the right number of
      -- edges is drawn.
      while 4-tile_edges ~= edges do
	 r_tile = math.random(#tiles)
	 tile_edges = #tiles[r_tile]
      end
      -- Check how much we need to rotate the tile
      if (row == 1 and col == 1) or (row == 1 and col == 2)  then
	 -- Top left corner or top middle
	 rotations = 3
      elseif (row == 1 and col == 3) or (row == 2 and col == 3) then
	 -- Top right corner or middle right
	 rotations = 2
      elseif (row == 3 and col == 2) or (row == 3 and col == 3) then
	 rotations = 1
      elseif (row == 2 and col == 2) then
	 -- Center tile
	 rotations = math.random(4)
      else
	 rotations = 0
      end
      -- Tables inside other tables are just references, so I make
      -- copies so that identical edge and corner tiles don't rotate together
      table.insert(board_tiles,deepcopy(tiles[r_tile]))
      -- Rotate the tiles.
      rotate_tile(rotations,board_tiles[#board_tiles])
   end
end

-- Check that nothing went badly.
assert(n_tiles == #board_tiles, "Number of assigned tiles does not match the desired board size")

-- At this point, board_tiles is the table which "is" the board.
-- Its elements are tables which represent the tiles.
-- The tile table elements are connections (1-4 of them) which are also tables
-- The connection tables are two-element, with the first being the tile edge
-- for the "incoming" (onto this tile) part of the connection, the second being
-- the tile edge for the "outgoing" part of the connection.

nodes = {}

Tile = class{
   init = function(self,idx,tile_table)
      self.idx = idx
      self.tile_table = tile_table
      self.pos_idx = tile_idx_to_idx_pos(idx)
      --self.n_edges = #tile_table
      self.connections = {}
      for i = 1,#self.tile_table do
	 local incoming = tile_table[i][1]
	 local outgoing = tile_table[i][2]
	 self.connections[incoming] = outgoing
	 table.insert(nodes,Node(idx,incoming,outgoing))
      end
   end
	    }

Board = class{
   init = function(self)
      self.tiles = {}
   end
	     }

Node = class{
   init = function(self,tile_idx,edge,next,previous)
      self.next = next
      self.previous = previous
      self.tile_idx = tile_idx
      self.edge = edge
   end,
   __tostring = function(self)
      return tostring(self.tile_idx)..":"..tostring(self.edge).."->"..tostring(self.next).." previous = "..tostring(self.previous)
   end
	    }

function tile_idx_to_idx_pos(tile_idx)
   return vector(((tile_idx - 1) % board_size.x)+1, 
		 math.floor((tile_idx+board_size.x-1)/board_size.x))
end

board = Board()
for i,tile in ipairs(board_tiles) do
   pos_idx = tile_idx_to_idx_pos(i)
   board.tiles[tostring(pos_idx)] = Tile(i,tile)
end

function adjacent_tile(pos_idx,edge)
   local shift = vector(0,0)
   if edge == 1 then
      shift.x = 1
   elseif edge == 2 then
      shift.y = -1
   elseif edge == 3 then
      shift.x = -1
   else
      shift.y = 1
   end
   return (pos_idx+shift)
end

-- print("List of nodes:")
-- for i,n in ipairs(nodes) do
--    print(n)
-- end

-- Link up the nodes (currently their "next" members just point to an edge number).
-- I would like to node.next member to be the next node in the chain.
-- Iterate over each node
for i,n in ipairs(nodes) do
   local this_pos_idx = tile_idx_to_idx_pos(n.tile_idx)
   local next_pos_idx = adjacent_tile(this_pos_idx,n.next)
   local next_idx = next_pos_idx.x+((next_pos_idx.y-1)*board_size.x)
   -- print(tostring(n))
   -- print(this_pos_idx)
   -- print(next_pos_idx)
   -- print(next_idx)
   for j,m in ipairs(nodes) do
      -- Loop over all the nodes again, looking for another node on the same tile
      -- whose edge is the target of the current one.
      if m.tile_idx == next_idx and (n.next-2)%4 == (m.edge)%4 then
	 -- Once found, make the next and previous members be the actual nodes
	 -- and break out of the loop.
	 nodes[i].next = nodes[j]
	 nodes[j].previous = nodes[i]
	 --print("Connecting nodes"..to_string(m).." and "..to_string(n))
	 --print("Connecting nodes "..tostring(i).." and "..tostring(j))
	 break
      end
   end
end

function Node:__tostring()
   return tostring(self.tile_idx)..":"..tostring(self.edge).."->"..tostring(self.next.tile_idx)..":"..tostring(self.next.edge)
end

-- print("List of nodes:")
-- for i,n in ipairs(nodes) do
--    print(n)
-- end

paths = {}
-- nodes will be emptied as I build the paths.
while #nodes > 0 do
   --print("Length of nodes: ",#nodes)
   local newpath = {}
   local node_idx = 1 -- I could use math.random(#nodes) here, but it doesn't matter.
   local firstnode = nodes[node_idx]
   local oldnode = nodes[node_idx]
   local newnode = nodes[node_idx]
   --print("firstnode:",firstnode)
   repeat 
      -- This block repeats until the loop is closed.
      table.insert(newpath,newnode)
      oldnode = newnode
      newnode = newnode.next -- Walk down the path...
      --print("newnode: ",newnode)
      -- Now I delete from nodes the node that was just inserted into the path.
      -- I do this by searching nodes until I find the one I want to delete.
      for i,n in ipairs(nodes) do
	 --print(n)
	 if n == oldnode then
	    --print("Removing node ",n," from list")
	    table.remove(nodes,i)
	    -- The break statement is not necessary, because there should only
	    -- be one of each node, but it allows the use of the assertion below.
	    break
	 end
	 -- This assertion should never be reached if a node was deleted.
	 --assert(false,"Error, could not find node "..tostring(oldnode).." to remove!")
      end
   until newnode == firstnode
   table.insert(paths,newpath)
   --print("Length of newpath:",#newpath)
end

--print(#paths)
--print(to_string(paths[1]))